# -*- coding: utf-8 -*-
#
# stdlib imports
#
import csv
import json
import logging
from pathlib import Path
#
# third-party imports
#
import numpy as np
import pandas as pd
#
# module imports
#
from .common import *
#
# read values from the log file and make optional plots
#
def process_logs(stats, logger):
    #
    # write overall stats
    #
    with (LOG_PATH / (PROGRAM_NAME + '_stats.tsv')).open('w', newline='') as resultfh:
        writer = csv.writer(resultfh, delimiter='\t', quoting=csv.QUOTE_MINIMAL)
        writer.writerow(['#'+ STAT_COLS[1]] + list(STAT_COLS[2:]) + [STAT_COLS[0]])
        for row in stats:
            writer.writerow(list(row[1:]) + [row[0]])
    #
    # get path to logfile
    #
    logfile_names = []
    for handler in logger.handlers:
        try:
            logfile_names.append(handler.stream.name)
        except:
            pass
    logfile_names.remove('<stderr>')
    logfile_path = Path(logfile_names[0])
    #
    # Shut down logging and parse log file
    #
    logging.shutdown()
    stat_dict = {}
    graph_dict = {}
    synonym_dict = {}
    for stat_name in STAT_TYPES:
        stat_dict[stat_name] = []
    for graph_name in GRAPH_TYPES:
        graph_dict[graph_name] = []
    tmpfile_path = logfile_path.parent/(logfile_path.name + 'tmp')
    with tmpfile_path.open('w') as tmpfh:
        with logfile_path.open('rU') as logfh:
            line = logfh.readline()
            while line:
                line = logfh.readline()
                parts = line.split('\t')
                if len(parts) > 2 and parts[2] in (GRAPH_TYPES + STAT_TYPES):
                    file = parts[0].strip('DEBUG: ')
                    record = parts[1]
                    rec_type = parts[2]
                    if rec_type in STAT_TYPES:
                        if rec_type in FLOAT_TYPES:
                            value = float(parts[3])
                        else:
                            value = int(parts[3])
                        stat_dict[rec_type].append((value, record, file))
                    else:
                        id = parts[3].strip()
                        if file not in synonym_dict:
                            synonym_dict[file] = {}
                        synonym_dict[file][record] = id
                        graph_dict[rec_type].append((file, id, record))
                else:
                    tmpfh.write(line)
            logfh.close()
    tmpfile_path.rename(logfile_path)
    for stat_name in STAT_TYPES:
        stat_frame = pd.DataFrame(stat_dict[stat_name],
                                  columns=['value', 'record', 'file'])
        stat_len = len(stat_frame)
        if stat_len > 0:
            stat_frame = stat_frame.sort_values(['value', 'record','file'])
            stat_frame.to_csv(LOG_PATH/(stat_name.rstrip('%')+'.tsv'),
                              sep='\t',
                              index=False,
                              header=['#'+stat_name,'id', 'file'])
            if stat_len > MIN_HIST_LEN:
                dist = np.array(stat_frame['value'])
                make_histogram(dist, stat_name, log10=True)
    for graph_name in GRAPH_TYPES:
        graph_frame = pd.DataFrame(graph_dict[graph_name],
                                   columns=['file','id','duplicate'])
        graph_len = len(graph_frame)
        if graph_len > 0:
            graph_frame = graph_frame.sort_values(['file', 'id', 'duplicate']
                                                  ).reindex(index=list(range(graph_len)))
            graph_frame.to_csv(LOG_PATH/(graph_name+'.tsv'),
                               sep='\t',
                               index=False,
                               header=['#file', 'id', graph_name])
    if len(synonym_dict):
        with (LOG_PATH/'synonyms.json').open('w') as synonym_fh:
            json.dump(synonym_dict, synonym_fh)



# -*- coding: utf-8 -*-
"""alphabetsoup -- fix alphabet and other problems in protein FASTA files.

Here are some common problems in protein-coding sequences and fixes:
Presence of stop codons at end - strip
Ambiguous residues at end - strip
Codes other than IUPAC + 'X' elsewhere - change to X
Length shorter than MINLEN - delete whole entry
"""
#
# standard library imports
#
import functools
import locale
import logging
import sys
import warnings
from datetime import datetime
#
# third-party imports
#
import click
import coverage
import dask.bag as db
import pandas as pd
from dask.diagnostics import ProgressBar
#
# package imports
#
from .version import version as VERSION
from .process_file import process_file
from .process_logs import process_logs
from .common import *
#
# Start coverage
#
coverage.process_startup()
# set locale so grouping works
for localename in ['en_US', 'en_US.utf8', 'English_United_States']:
    try:
        locale.setlocale(locale.LC_ALL, localename)
        break
    except:
        continue

AUTHOR = 'Joel Berendzen'
EMAIL = 'joelb@ncgr.org'
COPYRIGHT = """Copyright (C) 2018, The National Center for Genome Resources.  All rights reserved.
"""
PROJECT_HOME = 'https://github.com/ncgr/alphabetsoup'

DEFAULT_FILE_LOGLEVEL = logging.DEBUG
DEFAULT_STDERR_LOGLEVEL = logging.INFO
DEFAULT_FIRST_N = 0 # only process this many files
STARTTIME = datetime.now()
DEFAULT_MINLEN = 0 # minimum gene size in residues
DEFAULT_MINSEQS = 0 # minimum number of sequences per file
DEFAULT_MAXAMBIG = 0.0 # maximum number of ambiguous characters per gene, (0 = all)
DEFAULT_FILETYPES = ('*.faa', '*.fa', '*.fasta')
#
# global logger object
#
logger = logging.getLogger(PROGRAM_NAME)
#
# private context function
#
_ctx = click.get_current_context
#
# Classes
#
class CleanInfoFormatter(logging.Formatter):
    def __init__(self, fmt = '%(levelname)s: %(message)s'):
        logging.Formatter.__init__(self, fmt)

    def format(self, record):
        if record.levelno == logging.INFO:
            return record.getMessage()
        return logging.Formatter.format(self, record)


#
# Helper functions
#
def init_dual_logger(file_log_level=DEFAULT_FILE_LOGLEVEL,
                     stderr_log_level=DEFAULT_STDERR_LOGLEVEL):
    '''Log to stderr and to a log file at different levels
    '''
    def decorator(f):
        @functools.wraps(f)
        def wrapper(*args, **kwargs):
            global logger
            # find out the verbose/quiet level
            if _ctx().params['verbose']:
                _log_level = logging.DEBUG
            elif _ctx().params['quiet']:
                _log_level = logging.ERROR
            else:
                _log_level = stderr_log_level
            logger.setLevel(file_log_level)
            stderrHandler = logging.StreamHandler(sys.stderr)
            stderrFormatter = CleanInfoFormatter()
            stderrHandler.setFormatter(stderrFormatter)
            stderrHandler.setLevel(_log_level)
            logger.addHandler(stderrHandler)
            if _ctx().params['log']: # start a log file in LOG_PATH
                logfile_path = LOG_PATH / (PROGRAM_NAME + '.log')
                if not LOG_PATH.is_dir():  # create LOG_PATH
                    try:
                        logfile_path.parent.mkdir(mode=0o755, parents=True)
                    except OSError:
                        logger.error('Unable to create log directory "%s"',
                                     logfile_path.parent)
                        raise OSError
                else:
                    if logfile_path.exists():
                        try:
                            logfile_path.unlink()
                        except OSError:
                            logger.error('Unable to remove existing log file "%s"',
                                         logfile_path)
                            raise OSError
                logfileHandler = logging.FileHandler(str(logfile_path))
                logfileFormatter = logging.Formatter('%(levelname)s: %(message)s')
                logfileHandler.setFormatter(logfileFormatter)
                logfileHandler.setLevel(file_log_level)
                logger.addHandler(logfileHandler)
            logger.debug('Command line: "%s"', ' '.join(sys.argv))
            logger.debug('%s version %s', PROGRAM_NAME, VERSION)
            logger.debug('Run started at %s', str(STARTTIME)[:-7])
            return f(*args, **kwargs)
        return wrapper
    return decorator


def init_user_context_obj(initial_obj=None):
    '''Put info from global options into user context dictionary
    '''
    def decorator(f):
        @functools.wraps(f)
        def wrapper(*args, **kwargs):
            global config_obj
            if initial_obj is None:
                _ctx().obj = {}
            else:
                _ctx().obj = initial_obj
            ctx_dict = _ctx().obj
            if _ctx().params['verbose']:
                ctx_dict['logLevel'] = 'verbose'
            elif _ctx().params['quiet']:
                ctx_dict['logLevel'] = 'quiet'
            else:
                ctx_dict['logLevel'] = 'default'
            for key in ['progress', 'first_n']:
                ctx_dict[key] = _ctx().params[key]
            return f(*args, **kwargs)
        return wrapper
    return decorator


@click.command(epilog=AUTHOR+' <'+EMAIL+'>.  '+COPYRIGHT)
@click.option('-v', '--verbose', is_flag=True, show_default=True,
              default=False, help='Debug info to stderr.')
@click.option('-q', '--quiet', is_flag=True, show_default=True,
              default=False, help='Suppress logging to stderr.')
@click.option('--progress', is_flag=True, show_default=True,
              default=False, help='Show a progress bar.')
@click.option('--first_n', default=DEFAULT_FIRST_N,
               help='Process only this many files. [default: all]')
@click.option('--minlen', default=DEFAULT_MINLEN,show_default=True,
               help='Minimum sequence length.')
@click.option('--minseqs', default=DEFAULT_MINSEQS,show_default=True,
               help='Minimum sequences in file.')
@click.option('--maxambig', default=DEFAULT_MAXAMBIG,
               help='Max fraction ambiguous. [default:any]')
@click.option('--log/--no-log', is_flag=True, show_default=True,
              default=True, help='Write analysis in ./' + LOG_DIR +'.')
@click.option('--lengths/--no-lengths', is_flag=True,
              show_default=True, default=True, help='Compute lengths.')
@click.option('--dedup/--no-dedup', is_flag=True, show_default=True,
              default=False, help='De-duplicate sequences.')
@click.option('--defrag/--no-defrag', is_flag=True, show_default=True,
              default=False, help='Remove exact substrings.')
@click.option('--stripdash/--no-stripdash', is_flag=True, show_default=True,
              default=True, help='Remove "-" characters.')
@click.option('--overwrite/--no-overwrite', is_flag=True, show_default=True,
              default=False, help='Overwrite input files.')
@click.argument('in_path', type=click.Path(exists=True,
                                           writable=True,
                                           resolve_path=True,
                                           allow_dash=True,
                                           ))
@click.version_option(version=VERSION, prog_name=PROGRAM_NAME)
@init_dual_logger()
@init_user_context_obj()
def cli(in_path,
        verbose,
        quiet,
        progress,
        first_n,
        log,
        overwrite,
        minlen,
        minseqs,
        maxambig,
        dedup,
        stripdash,
        defrag,
        lengths):
    """alphabetsoup -- fix alphabet and other problems in protein FASTA files
    """
    if quiet or verbose:
        progress = False
    if defrag:
        dedup = True
    if not verbose:
        warnings.filterwarnings('ignore')

    logger.debug('')
    in_path = Path(in_path)
    files = []
    for ext in DEFAULT_FILETYPES:
        files.extend(in_path.rglob(ext))
    if first_n > 0 and len(files) > first_n:
        files = files[:first_n]
    if progress and not quiet:
        logger.info('Processing %d files in parallel:', len(files))
        ProgressBar().register()
    bag = db.from_sequence(files)
    results = bag.map(process_file,
                      logger=logger,
                      write=overwrite,
                      min_len=minlen,
                      min_seqs=minseqs,
                      max_ambiguous=maxambig,
                      remove_duplicates=dedup,
                      remove_dashes=stripdash,
                      remove_substrings=defrag,
                      lengths=lengths).compute()
    if log:
        if not quiet:
            logger.info('Processing log files serially:')
        process_logs(results, logger)
    #
    # Transpose results and put Name as first column
    #
    max_qty_len = max([len(s) for s in STAT_COLS])
    df = pd.DataFrame(results,
                      columns=STAT_COLS)
    if not quiet:
        print('Quantity\tSum\tMean')
        print('%s:\t%d\t0' %('Files'.rjust(max_qty_len), len(files)))

    for stat_name in STAT_COLS[1:]:
        dist = df[stat_name]
        stat_len = len(dist)
        mean = dist.mean()
        if not quiet:
            print('%s:\t%d\t%.3f'%(stat_name.rjust(max_qty_len),
                               dist.sum(),
                               mean))
        if log and max(dist)>0 and stat_len > MIN_HIST_LEN:
            make_histogram(dist, stat_name)
    sys.exit(0)



# -*- coding: utf-8 -*-
# stdlib imports
import zlib
from collections import defaultdict
from operator import itemgetter
# third-party imports
from Bio import SeqIO
from Bio.Data import IUPACData
# package imports
from .common import *

ALPHABET = IUPACData.protein_letters + 'X' + '-'
LOGINT_FMT = '%s\t%s\t%s\t%d'
LOGFLOAT_FMT = '%s\t%s\t%s\t%f'
LOGSTR_FMT = '%s\t%s\t%s\t%s'

class DuplicateIDDict(defaultdict):
    """A dictionary of lists with a get() that returns a mangled ID
    """
    def __init__(self):
        super().__init__(list)

    def get(self,k):
        return '|'.join([k] + self[k])


def process_file(file,
                 logger=None,
                 write=False,
                 min_len=0,
                 min_seqs=0,
                 max_ambiguous=0,
                 remove_duplicates=False,
                 remove_dashes=False,
                 remove_substrings=False,
                 lengths=False):
    if logger:
        logger.debug('processing %s', file)
    out_sequences = []
    out_len = 0
    n_ambig = 0
    seq_hash_dict = {}
    n_dups = 0
    n_substrings = 0
    n_seqs_in = 0
    n_short = 0
    duplicate_ID_dict = DuplicateIDDict()
    with file.open('rU') as handle:
        for record in SeqIO.parse(handle, 'fasta'):
            n_seqs_in += 1
            seq = record.seq.upper().tomutable()
            if remove_dashes:
                # delete '-' as insertion characters in an alignment
                [ seq.pop(dash_pos-k) for k, dash_pos in
                    enumerate([i for i,j in enumerate(seq) if j == '-'])]
            # replace everything else out of alphabet with 'X'-
            [seq.__setitem__(i, 'X')
             for i, j in enumerate(seq) if j not in ALPHABET]
            # remove trailing ambiguous/stop
            while seq[-1] == 'X':
                seq.pop()
            # remove leading ambiguous
            while seq[0] == 'X':
                seq.pop(0)
            # discard if too short
            if len(record.seq) < min_len:
                n_short += 1
                if logger:
                    logger.debug(LOGINT_FMT,
                                 file.name,
                                 record.id,
                                 SHORT_NAME,
                                 len(record.seq))
                continue
            # count duplicates and optionally discard
            seq_hash = zlib.adler32(bytearray(str(seq),'utf-8'))
            if seq_hash not in seq_hash_dict:
                seq_hash_dict[seq_hash] = record.id
            else:
                n_dups += 1
                first_ID = seq_hash_dict[seq_hash]
                duplicate_ID_dict[first_ID].append(record.id)
                if remove_duplicates:
                    logger.debug(LOGSTR_FMT,
                                 file.name,
                                 record.id,
                                 DUP_NAME,
                                 first_ID)
                    continue
            # count interior X's and discard if more than max_ambiguous
            ambig = sum([i =='X' for i in seq])
            if ambig > max_ambiguous:
                n_ambig += 1
                if logger:
                    logger.debug(LOGFLOAT_FMT,
                                 file.name,
                                 record.id,
                                 AMBIG_NAME,
                                 ambig*100./len(seq))
                continue
            record.seq = seq.toseq()
            out_sequences.append(record)
    # Search for exact substring matches in the set
    length_idx = [(i,len(record.seq))
                  for i, record in enumerate(out_sequences)]
    length_idx.sort(key=itemgetter(1))
    ascending = [idx for idx,length in length_idx]
    subst_removal_list = []
    for item_num,idx in enumerate(ascending):
        test_seq = out_sequences[idx].seq
        test_id = out_sequences[idx].id
        # traverse from biggest to smallest to find the biggest match
        for record in [out_sequences[i] for i in reversed(ascending[item_num+1:])]:
            if str(test_seq) in str(record.seq):
                n_substrings += 1
                duplicate_ID_dict[record.id].append(test_id)
                subst_removal_list.append(idx)
                if logger:
                    logger.debug(LOGSTR_FMT,
                                 file.name,
                                 test_id,
                                 SUBSTRING_NAME,
                                 record.id)
                break
    # Optionally remove exact substring matches
    if remove_substrings and len(subst_removal_list) > 0:
        subst_removal_list.sort()
        for item_num, idx in enumerate(subst_removal_list):
            out_sequences.pop(idx-item_num)
    # Count length of output residues
    for record in out_sequences:
        out_len += len(record.seq)
    # Indicate (in ID) records deduplicated
    if (remove_duplicates or remove_substrings) and len(duplicate_ID_dict) > 0:
        for record in out_sequences:
            if record.id in duplicate_ID_dict:
                record.id = duplicate_ID_dict.get(record.id)
                record.description = ''
    # If file is small, log it.  Else write lengths if requested.
    if len(out_sequences) < min_seqs:
        logger.debug(LOGINT_FMT,
                     file.name,
                     record.id,
                     FILESMALL_NAME,
                     len(out_sequences))
        small = 1
    else:
        if lengths and logger:
            for record in out_sequences:
                logger.debug(LOGINT_FMT,
                             file.name,
                             record.id,
                             LENGTH_NAME,
                             len(record.seq))
        small = 0
    if write:
        if not small:
            with file.open('w') as output_handle:
                SeqIO.write(out_sequences,output_handle,'fasta')
        else:
            logger.debug('file %s has only %d sequences after processing, removed',
                        file.name, len(out_sequences))
            file.unlink()
    return (file.name,
            n_seqs_in,
            len(out_sequences),
            out_len,
            n_ambig,
            n_short,
            small,
            n_dups,
            n_substrings)



# -*- coding: utf-8 -*-
# imports
from pathlib import Path
import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns
# global variables
PROGRAM_NAME = 'alphabetsoup'
DUP_NAME = 'Dups'
SUBSTRING_NAME = 'Substr'
GRAPH_TYPES = [DUP_NAME, SUBSTRING_NAME]
AMBIG_NAME = 'Ambig%'
SHORT_NAME = 'Short'
FILESMALL_NAME = 'Small'
INT_TYPES = [SHORT_NAME, FILESMALL_NAME]
FLOAT_TYPES = [AMBIG_NAME]
NUMBER_TYPES = FLOAT_TYPES + INT_TYPES
STRING_TYPES = ['Name']
LENGTH_NAME = 'Length'
LOG_TYPES = [LENGTH_NAME]
STAT_TYPES = NUMBER_TYPES + LOG_TYPES
STAT_COLS = ['name',
             'seqs_in',
             'seqs_out',
             'residues',
             'n_ambig',
             'n_short',
             'n_small',
             'n_dups',
             'n_substr']
MIN_HIST_LEN = 50
PLOT_TYPES = ['png','svg']
LOG_DIR = 'log'
LOG_PATH = Path('.')/ LOG_DIR

def make_histogram(dist, name, log10=False):
    # do histogram plot with kernel density estimate
    mean = dist.mean()
    if log10:
        dist = np.log10(dist)
    sns.distplot(dist,
                 rug=True,
                 rug_kws={'color': 'b'},
                 kde_kws={'color': 'k',
                          'linewidth': 1,
                          'label': 'KDE'},
                 hist_kws={'histtype': 'step',
                           'linewidth': 2,
                           'alpha': 1,
                           'color': 'b'}
                 )
    plt.title('%s histogram of %d values, mean=%.1f'
              % (name, len(dist), mean))
    if log10:
        plt.xlabel('log ' + name)
    else:
        plt.xlabel(name)
    plt.ylabel('Frequency')
    for ext in PLOT_TYPES:
        plt.savefig(LOG_PATH / ('%s-histogram.' % (name.rstrip('%')) + ext),
                    bbox_inches='tight')
    plt.close('all')